@@ -1,6 +1,9 @@
 use crate::tokenizer::{NormalizedString, Normalizer, Result};
+use serde::{Deserialize, Serialize};
 
+#[derive(Serialize, Deserialize)]
 pub struct NFD;
+#[typetag::serde]
 impl Normalizer for NFD {
     fn normalize(&self, normalized: &mut NormalizedString) -> Result<()> {
         normalized.nfd();
@@ -8,7 +11,9 @@ impl Normalizer for NFD {
     }
 }
 
+#[derive(Serialize, Deserialize)]
 pub struct NFKD;
+#[typetag::serde]
 impl Normalizer for NFKD {
     fn normalize(&self, normalized: &mut NormalizedString) -> Result<()> {
         normalized.nfkd();
@@ -16,7 +21,9 @@ impl Normalizer for NFKD {
     }
 }
 
+#[derive(Serialize, Deserialize)]
 pub struct NFC;
+#[typetag::serde]
 impl Normalizer for NFC {
     fn normalize(&self, normalized: &mut NormalizedString) -> Result<()> {
         normalized.nfc();
@@ -24,7 +31,9 @@ impl Normalizer for NFC {
     }
 }
 
+#[derive(Serialize, Deserialize)]
 pub struct NFKC;
+#[typetag::serde]
 impl Normalizer for NFKC {
     fn normalize(&self, normalized: &mut NormalizedString) -> Result<()> {
         normalized.nfkc();
