@@ -1,9 +1,10 @@
+use crate::parallelism::*;
 use crate::tokenizer::{Offsets, Token};
 use crate::utils::padding::PaddingDirection;
-use rayon::prelude::*;
+use serde::{Deserialize, Serialize};
 
 /// Represents the output of a `Tokenizer`.
-#[derive(Default, PartialEq, Debug, Clone)]
+#[derive(Default, PartialEq, Debug, Clone, Serialize, Deserialize)]
 pub struct Encoding {
     /// IDs produced by the `Tokenizer`
     ids: Vec<u32>,
@@ -76,6 +77,14 @@ impl Encoding {
         }
     }
 
+    pub fn is_empty(&self) -> bool {
+        self.ids.is_empty()
+    }
+
+    pub fn len(&self) -> usize {
+        self.ids.len()
+    }
+
     pub fn get_tokens(&self) -> &[String] {
         &self.tokens[..]
     }
@@ -84,6 +93,10 @@ impl Encoding {
         &self.words
     }
 
+    pub fn get_words_mut(&mut self) -> &mut [Option<u32>] {
+        &mut self.words
+    }
+
     pub fn get_ids(&self) -> &[u32] {
         &self.ids
     }
@@ -349,7 +362,7 @@ impl Encoding {
         direction: PaddingDirection,
     ) {
         // Dispatch call to all the overflowings first
-        self.overflowing.par_iter_mut().for_each(|encoding| {
+        self.overflowing.maybe_par_iter_mut().for_each(|encoding| {
             encoding.pad(target_length, pad_id, pad_type_id, pad_token, direction)
         });
 
